#include <linux/version.h>
#include <linux/device.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/cdev.h>
#include <linux/fs.h>

#include <asm/uaccess.h>

#include <rtdm/rtdm_driver.h>

#define nom_device_rtdm "mesure_periode_rtdm"


static int read_mesure_periode(struct rtdm_dev_context * contexte,
                               rtdm_user_info_t * info, 
                               void * buffer, size_t lg);
                        
static int open_nonrt_mesure_periode (struct rtdm_dev_context * contexte, 
                                      rtdm_user_info_t * info,
                                      int flags);

static int close_nonrt_mesure_periode (struct rtdm_dev_context * contexte, 
                                       rtdm_user_info_t * info);

static void handler_mesure_periode (rtdm_timer_t * timer);


static struct rtdm_device rtdev_mesure_periode = {
	.struct_version = RTDM_DEVICE_STRUCT_VER,
	.device_flags   = RTDM_NAMED_DEVICE,
	.context_size   = 0,
	.device_name    = nom_device_rtdm,
	.open_nrt       = open_nonrt_mesure_periode,
	.ops = {
		.close_nrt = close_nonrt_mesure_periode,
		.read_nrt  = read_mesure_periode,
	},
	.device_class     = RTDM_CLASS_TESTING,
	.device_sub_class = 1,
	.profile_version  = 1,
	.driver_name      = nom_device_rtdm,
	.driver_version   = RTDM_DRIVER_VER(1,0,0),
	.peripheral_name  = nom_device_rtdm,
	.provider_name    = "cpb",
	.proc_name        = nom_device_rtdm,
};


static nanosecs_abs_t  periode_mini;
static nanosecs_abs_t  periode_maxi;
static nanosecs_abs_t  somme_periodes;
static int             nb_periodes;
static rtdm_mutex_t mtx_statistiques;

static rtdm_timer_t rtimer;
static nanosecs_rel_t periode_desiree = 1000000; // 1 ms

static int __init init_mesure_periode (void)
{
	int err;

	rtdm_mutex_init(& mtx_statistiques);
	periode_mini = -1;
	periode_maxi = 0;
	somme_periodes = 0;
	nb_periodes = 0;

	rtdm_dev_register (& rtdev_mesure_periode);
	if ((err = rtdm_timer_init(& rtimer, handler_mesure_periode, "Timer_mesure_periode")) != 0) {
		rtdm_dev_unregister(& rtdev_mesure_periode, 0);
		rtdm_mutex_destroy(& mtx_statistiques);
		return err;
	}

	if ((err = rtdm_timer_start(& rtimer, 1000000000, periode_desiree, RTDM_TIMERMODE_RELATIVE)) != 0) {
		rtdm_timer_destroy(& rtimer);
		rtdm_dev_unregister(& rtdev_mesure_periode, 0);
		rtdm_mutex_destroy(& mtx_statistiques);
		return err;
	}
	
	return 0; 
}



static void __exit exit_mesure_periode (void)
{
	rtdm_timer_stop(& rtimer);
	rtdm_timer_destroy(& rtimer);
	rtdm_dev_unregister (& rtdev_mesure_periode, 0);
	rtdm_mutex_destroy(& mtx_statistiques);
}



static int read_mesure_periode(struct rtdm_dev_context * contexte,
                               rtdm_user_info_t * info, 
                               void * buffer, size_t lg)
{
	char local_buffer[80];
	nanosecs_abs_t  _periode_mini;
	nanosecs_abs_t  _periode_maxi;
	nanosecs_abs_t  _somme_periodes;
	int             _nb_periodes;
	
	rtdm_mutex_lock(& mtx_statistiques);
	_periode_mini   = periode_mini;
	_periode_maxi   = periode_maxi;
	_somme_periodes = somme_periodes;
	_nb_periodes    = nb_periodes;
	rtdm_mutex_unlock(& mtx_statistiques);
	
	snprintf(local_buffer, 80, "%lld %lld %lld %d\n", 
		_periode_mini, _periode_maxi, _somme_periodes, _nb_periodes);
	if (rtdm_safe_copy_to_user(info, buffer, local_buffer, strlen(local_buffer)) != 0)
		return -EFAULT;
	return strlen(local_buffer);
}



static int open_nonrt_mesure_periode(struct rtdm_dev_context * contexte, 
                                     rtdm_user_info_t * info,
                                     int flags) 
{
	return 0;
}



static int close_nonrt_mesure_periode (struct rtdm_dev_context * contexte, 
                                       rtdm_user_info_t * info) 
{
	return 0;
}


static void handler_mesure_periode (rtdm_timer_t * timer)
{
	nanosecs_abs_t         heure;
	static nanosecs_abs_t  precedente;
	nanosecs_abs_t         periode;
	
	heure = rtdm_clock_read();
	rtdm_mutex_lock(& mtx_statistiques);
	if (nb_periodes > 0) {
		periode = heure - precedente;
		if (periode > periode_maxi)
			periode_maxi = periode;
		if ((periode_mini == -1) || (periode_mini > periode))
			periode_mini = periode;
		somme_periodes += periode;
	}
	nb_periodes ++;
	rtdm_mutex_unlock(& mtx_statistiques);
	precedente = heure;
}



module_init(init_mesure_periode);
module_exit(exit_mesure_periode);
MODULE_LICENSE("GPL");

